#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "camcap.h"

static void camcap_errno(char *text, int errnoval)
{
	fprintf(stderr, "camcap: %s - %s\n", text, strerror(errnoval));
}

/*
 * camcap_open
 *
 * camcap_open_t *co: set to NULL for now
*/
camcap_t *camcap_open(char *fname, camcap_open_t *co)
{
	camcap_t *cc = NULL;
	struct v4l2_cropcap cropcap;
	struct v4l2_crop crop;
	struct v4l2_format fmt;
	unsigned int min;

	if (!fname)
		goto errout;

	if ((cc = calloc(sizeof(camcap_t), 1)) == NULL) {
		camcap_errno("calloc", errno);
		goto errout;
	}

	cc->method = IOM_USERPTR;

	if ((cc->fname = malloc(strlen(fname)+1)) == NULL) {
		camcap_errno("malloc", errno);
		goto errout;
	}
	strcpy(cc->fname, fname);

	/* opening for O_RDWR instead of O_RDONLY because the
	   capture_example.c said it was required
	*/
	if ((cc->fd = open(cc->fname, O_RDWR | O_NONBLOCK, 0)) == -1) {
		camcap_errno(cc->fname, errno);
		goto errout;
	}

	if (ioctl(cc->fd, VIDIOC_QUERYCAP, &cc->cap) == -1) {
		camcap_errno("VIDIOC_QUERYTCAP", errno);
		goto errout; // EINVAL means not a V4L2 device
	}

	if (!(cc->cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)) {
		camcap_errno("V4L2_CAP_VIDEO_CAPTURE", ENOSYS);
		goto errout;
	}

	if ((cc->method == IOM_USERPTR /*|| cc->method == IOM_MMAP */) &&
		!(cc->cap.capabilities & V4L2_CAP_STREAMING)) {
		camcap_errno("V4L2_CAP_STREAMING", ENOSYS);
		goto errout;
	} /* else if (cc->method == IOM_READ &&
		!(cc->cap.capabilities & V4L2_CAP_READWRITE)) {
		errno = ENOSYS;
		goto errout;
	}*/ /* this method is not supported by the YoLuke and is not implemented */
	
	memset(&cropcap, 0, sizeof(cropcap));
	cropcap.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	if (ioctl(cc->fd, VIDIOC_CROPCAP, &cropcap) == 0) {
		crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		crop.c = cropcap.defrect; /* reset to default */
		ioctl(cc->fd, VIDIOC_S_CROP, &crop);
		/* if it returns -1 and errno == EINVAL then cropping not supported
	 	   but we don't care  */
	} /* else ignore errors */

	memset(&fmt, 0, sizeof(fmt));
	fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	if (ioctl(cc->fd, VIDIOC_G_FMT, &fmt) == -1) {
		camcap_errno("VIDIOC_G_FMT", errno);
		goto errout;
	}
	if (fmt.fmt.pix.pixelformat != V4L2_PIX_FMT_YUYV) {
		camcap_errno("Not V4L2_PIX_FMT_YUYV", ENOSYS);
		goto errout;
	}
	cc->width = fmt.fmt.pix.width;
	cc->height = fmt.fmt.pix.height;
	cc->stride = fmt.fmt.pix.bytesperline;
	cc->imagesize = fmt.fmt.pix.sizeimage;
	/* defense against some buggy drivers, copied from:
	 * https://www.kernel.org/doc/html/v4.15/media/uapi/v4l/capture-example.html
	*/
	/* GOBACK: The above sample also does defence for bytesperline */
	min = fmt.fmt.pix.bytesperline * cc->height;
	if (cc->imagesize < min)
		cc->imagesize = min;
	cc->pixelformat = fmt.fmt.pix.pixelformat;

	return cc;

errout:
	if (cc) {
		if (cc->fname)
			free(cc->fname);
		free(cc);
	}
	return NULL;
}

void camcap_get_sizes(camcap_t *cc, unsigned *width, unsigned *height, unsigned *imagesize)
{
	if (width) *width = cc->width;
	if (height) *height = cc->height;
	if (imagesize) *imagesize = cc->imagesize;
}

static int camcap_init_userptr(camcap_t *cc)
{
	struct v4l2_requestbuffers req;

	memset(&req, 0, sizeof(req));
	req.count = 1;
	req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	req.memory = V4L2_MEMORY_USERPTR;
	if (ioctl(cc->fd, VIDIOC_REQBUFS, &req) == -1) {
		camcap_errno("VIDIOC_REQBUFS", errno);
		return -1;
	}
	return 0;
}

/*
 * camcap_init
 *
 * **vb: If the cc->method is IOM_USERPTR then this should be a
 *       buffer for the video driver to fill, it should be
 *       cc->imagesize in bytes.
 *       If IOM_MMAP is ever supported then this will be used to
 *       instead return the video device buffer.
*/
int camcap_init(camcap_t *cc, void **vb)
{
	int rc;

	switch (cc->method) {
	case IOM_USERPTR:
		cc->vb = *vb;
		rc = camcap_init_userptr(cc);
		break;
	}
	return rc;
}

int camcap_start_capturing(camcap_t *cc)
{
	struct v4l2_buffer buf;
	enum v4l2_buf_type type;

	switch (cc->method) {
	case IOM_USERPTR:
		memset(&buf, 0, sizeof(buf));
		buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		buf.memory = V4L2_MEMORY_USERPTR;
		buf.index = 0;
		buf.m.userptr = (unsigned long) cc->vb;
		buf.length = cc->imagesize;
		if (ioctl(cc->fd, VIDIOC_QBUF, &buf) == -1) {
			camcap_errno("VIDIOC_QBUF", errno);
			return -1;
		}
		type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		if (ioctl(cc->fd, VIDIOC_STREAMON, &type) == -1) {
			camcap_errno("VIDIOC_STREAMON", errno);
			return -1;
		}
		break;
	}

	return 0;
}

int camcap_read_frame(camcap_t *cc)
{
	struct v4l2_buffer buf;

	switch (cc->method) {
	case IOM_USERPTR:
		memset(&buf, 0, sizeof(buf));
		buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		buf.memory = V4L2_MEMORY_USERPTR;
		if (ioctl(cc->fd, VIDIOC_DQBUF, &buf) == -1) {
			/* If opened for O_NONBLOCK then this may return EAGAIN. */
			/* Could ignore EIO as it can mean temporary loss of signal. */
			camcap_errno("VIDIOC_DQBUF", errno);
			return -1;
		}
		break;
	}
	/* GOBACK: in capture_example.c it next processes the image and then
       queues it up again. Consider having multiple buffers and letting
       the driver fill up another buffer while we process one. 
	   For now I have camcap_requeue_frame() below and leave it to the
	   caller.
	*/
	return 0;
}

int camcap_requeue_frame(camcap_t *cc)
{
	struct v4l2_buffer buf;

	memset(&buf, 0, sizeof(buf));
	buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	buf.memory = V4L2_MEMORY_USERPTR;
	buf.index = 0;
	buf.m.userptr = (unsigned long) cc->vb;
	buf.length = cc->imagesize;
	if (ioctl(cc->fd, VIDIOC_QBUF, &buf) == -1) {
		camcap_errno("VIDIOC_QBUF", errno);
		return -1;
	}
	return 0;
}

int camcap_stop_capturing(camcap_t *cc)
{
	enum v4l2_buf_type type;

	switch (cc->method) {
	case IOM_USERPTR:
		type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		if (ioctl(cc->fd, VIDIOC_STREAMOFF, &type) == -1) {
			camcap_errno("VIDIOC_STREAMOFF", errno);
			return -1;
		}
		break;
	}

	return 0;
}

void camcap_close(camcap_t *cc)
{
	if (cc) {
		if (cc->fname)
			free(cc->fname);
		free(cc);
	}
}
